/*
  plcLib Version 2.0 Beta 6, last updated 22nd April, 2023.
  
  Simple PLC-style programming for the Arduino and compatibles.

  Authors: 
  
      Walter Ditch - Version 0.5 to present
      Stevan Tosic - Version 2.0 to present

  Licence: MIT
  
  Copyright (c) 2023, the plcLib authors

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in all
  copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.
  
*/

#ifndef plcLib_h
#define plcLib_h

#include "Arduino.h"

#ifndef noPinDefs
/*
Define basic I/O pins for Arduino Uno / Arduino MKR and compatibles
(or leave default I/O pins unconfigured if noPinDefs is set in the user sketch)
*/

// Default X0/X1 connections for Arduino Uno + Grove Base Shield

const int X0 = 7;   // D0   Din   Push button
const int X1 = 8;   // D1   Din   Push button

// Alternative X0/X1 connections for Arduino MKR + Grove Connector Carrier]
// (For MKR, comment out previous 2 lines and remove comments on 2 lines below)
// const int X0 = 0;   // D0   Din   Push button
// const int X1 = 1;   // D1   Din   Push button

const int X2 = A2;  // A2   Din   Push button
const int X3 = A3;  // A3   Din   Push button

const int Y0 = 2;   // D2~  out  LED
const int Y1 = 3;   // D3~~ Dout  LED
const int Y2 = 4;   // D4~  Dout  LED
const int Y3 = 5;   // D5~~ Dout  LED

const int AD0 = A0; // Analogue input 0
const int AD1 = A1; // Analogue input 1

//  ~~ = PWM compatible (both), ~ = MKR only

#endif /* noPinDefs section is undefined */

// Constants used when definining pin types (Pin class)
const int DIGITAL = 0;  // 0
const int ANALOGUE = 1;  // 1 (UK)
const int ANALOG = 1;    // 1

extern unsigned long scanValue;

void setupPLC();
unsigned long digitalRead2(int input);
void digitalWrite2(int output, unsigned int value);
unsigned long din(int input);
unsigned long din(unsigned long input);
unsigned long dinNot(int input);
unsigned long dinNot(unsigned long input);
unsigned long ain(int input);
unsigned long ain(unsigned long input);
unsigned long dout(int output);
unsigned long dout(unsigned long &output);
unsigned long doutNot(int output);
unsigned long doutNot(unsigned long &output);
unsigned long pout(int output);
unsigned long andBit(int input);
unsigned long andBit(unsigned long input);
unsigned long andNotBit(int input);
unsigned long andNotBit(unsigned long input);
unsigned long orBit(int input);
unsigned long orBit(unsigned long input);
unsigned long orNotBit(int input);
unsigned long orNotBit(unsigned long input);
unsigned long xorBit(int input);
unsigned long xorBit(unsigned long input);
unsigned long xorNotBit(int input);
unsigned long xorNotBit(unsigned long input);
unsigned long latch(int output, int reset);
unsigned long latch(int output, unsigned long reset);
unsigned long latch(unsigned long &output, int reset);
unsigned long latch(unsigned long &output, unsigned long reset);
unsigned long timerOn(unsigned long &timerState, unsigned long timerPeriod);
unsigned long timerPulse(unsigned long &timerState, unsigned long timerPeriod);
unsigned long timerOff(unsigned long &timerState, unsigned long timerPeriod);
unsigned long compareGT(int input);
unsigned long compareGT(unsigned long input);
unsigned long compareLT(int input);
unsigned long compareLT(unsigned long input);
unsigned long setL(int output);
unsigned long setL(unsigned long &output);
unsigned long resetL(int output);
unsigned long resetL(unsigned long &output);

class Object
{
  public:
    Object();
    Object(unsigned long val);
	virtual unsigned long value();
	virtual unsigned long setValue(unsigned long val);

  protected:
	unsigned long _val;
};

class Pin : public Object
{
  public:
    // Pin();
    // Pin(int id);
    // Pin(int id, int mode, int type = 0);
    Pin(int id, int mode, int type);

	unsigned long value();
	unsigned long setValue(unsigned long val);
    unsigned long mode();
    unsigned long type();
  
  private:      // _val inherited from Object
    int _id;    // Pin number
    int _mode;  // 0 = INPUT, 1 = OUTPUT, 2 = INPUT_PULLUP
    int _type;  // 0 = digital, 1 = analogue, (others reserved)
};

class Counter : public Object
{
  public:
    Counter(unsigned long presetValue);
    Counter(unsigned long presetValue, unsigned long direction);
    void countUp();
    void countDown();
    void preset();
    void clear();
    unsigned long upperQ();
    unsigned long lowerQ();
    unsigned long value();
    unsigned long setValue(unsigned long val);
  private:
    unsigned long _pv;
    unsigned long _ctUpEdge;
	unsigned long _ctDownEdge;
    unsigned long _uQ;
    unsigned long _lQ;
};

class Shift : public Object
{
  public:
    Shift();
	Shift(unsigned long val);
	unsigned long bitValue(unsigned int bitno);
	unsigned long value();
    unsigned long setValue(unsigned long val);
	void reset();
	void inputBit();
	void shiftLeft();
	void shiftRight();
  private:
    unsigned long _srLeftEdge;
    unsigned long _srRightEdge;
	unsigned long _inbit;
};

class Stack : public Object
{
  public:
    Stack();
	void push();
	void pop();
	void orBlock();
	void andBlock();
    unsigned long value();
};

class Pulse : public Object
{
  public:
    Pulse();
    Pulse(unsigned int polarity);
	void inClock();
	void rising();
	void falling();
    unsigned long value();
    void expired();

  private:
	unsigned int _pulseUpEdge;
	unsigned int _pulseDownEdge;
    unsigned int _polarity;
};

struct Waveform
{
  unsigned long elapsedLow;
  unsigned long elapsedHigh;
};

unsigned long timerCycle(Waveform &timerName, unsigned long timer1Period, unsigned long timer2Period);

// Function Block definitions, based on Templates

template<typename TFB, typename TIN>
unsigned long orFB(TFB &fbVar, TIN i)
{
    fbVar = din(i);
    scanValue = fbVar;
    return(scanValue);
}

// OR Function Block (multiple input, single bit)
template<typename TFB, typename TIN, typename... TINs>
unsigned long orFB(TFB &fbVar, TIN i, TINs... inputs)
{
    orFB(fbVar, inputs...);
    fbVar |= din(i);
    scanValue = fbVar;
    return(scanValue);
}

template<typename TFB, typename... TINs>
unsigned long norFBT(TFB &fbVar, TINs... inputs)
{
    orFBT(fbVar, inputs...);
    fbVar = ~fbVar & 1;
    scanValue = fbVar;
    return(scanValue);
}

// NOR Function Block  (multiple input, single bit)
template<typename TFB, typename... TINs>
unsigned long norFB(TFB &fbVar, TINs... inputs)
{
    orFB(fbVar, inputs...);
    fbVar = ~fbVar & 1;
    scanValue = fbVar;
    return(scanValue);
}

// XOR Function Block  (multiple input, single bit)
template<typename TFB, typename TIN>
unsigned long xorFB(TFB &fbVar, TIN i)
{
    fbVar = din(i);
    scanValue = fbVar;
    return(scanValue);
}

template<typename TFB, typename TIN, typename... TINs>
unsigned long xorFB(TFB &fbVar, TIN i, TINs... inputs)
{
    xorFB(fbVar, inputs...);
    fbVar ^= din(i);
    scanValue = fbVar;
    return(scanValue);
}


// XNOR Function Block  (multiple input, single bit)
template<typename TFB, typename... TINs>
unsigned long xnorFB(TFB &fbVar, TINs... inputs)
{
    xorFB(fbVar, inputs...);
    fbVar = ~fbVar & 1;
    scanValue = fbVar;
    return(scanValue);
}


// AND Function Block  (multiple input, single bit)
template<typename TFB, typename TIN>
unsigned long andFB(TFB &fbVar, TIN i)
{
    fbVar = din(i);
    scanValue = fbVar;
    return(scanValue);
}

template<typename TFB, typename TIN, typename... TINs> 
unsigned long andFB(TFB &fbVar, TIN i, TINs... inputs) 
{ 
    andFB(fbVar, inputs...);
    fbVar &= din(i);
    scanValue = fbVar;
    return(scanValue);
}

// NAND Function Block  (multiple input, single bit)
template<typename TFB, typename... TINs>
unsigned long nandFB(TFB &fbVar, TINs... inputs)
{
    andFB(fbVar, inputs...);
    fbVar = ~fbVar & 1;
    scanValue = fbVar;
    return(scanValue);
}

// Not Function Block (single input only, single bit)
template<typename TFB, typename TIN>
unsigned long notFB(TFB &fbVar, TIN i)
{
    fbVar = ~din(i) &1;
    scanValue = fbVar;
    return(scanValue);
}

class State {    

  private:
	unsigned long eEdge;           // Entering flag (entry event true for 1 scan cycle)
	unsigned long lEdge;           // Leaving flag (exit event true for 1 scan cycle)
	unsigned long _val;            // enable (1) / disable (0) / getValue = active
    
  public:

    State();
    State(unsigned long val);
    
    unsigned long value();
    unsigned long active();
    void enable() ;
    void disable();
    void back();
    unsigned long entry();
    unsigned long exit();
    
    State *history_ptr;            // A pointer to the previous state, if any (or NULL)
  
    //Properties copied from Pulse, renamed and made public
    unsigned int stateInput;       // Input tracker
	unsigned int stateUpEdge;      // Rising edge detect
	unsigned int stateDownEdge;    // Falling edge detect
};


// Digital I/O and single bit logical commands based on States
unsigned long din(State &s);
unsigned long dinNot(State &s);
unsigned long dout(State &s);
unsigned long doutNot(State &s);
unsigned long andBit(State &s);
unsigned long andNotBit(State &s);
unsigned long orBit(State &s);
unsigned long orNotBit(State &s);
unsigned long xorBit(State &s);
unsigned long xorNotBit(State& s);  // Name corrected 3/8/22 by WD

// New transition command (level based by default)
unsigned long trans(State& src, State& dst);   // Level-based transition (renamed from transL)


// Read an input supplied as Object
unsigned long din(Object &input);
unsigned long dinNot(Object &input);
unsigned long dout(Object &output);
unsigned long doutNot(Object &output);

using FunctionBlock = unsigned long;  // Create an alias so FunctionBlock equates to unsigned long
using Auxiliary = unsigned long;  // Create an alias so Auxiliary equates to unsigned long
using Timer = unsigned long;  // Create an alias so Timer equates to unsigned long


#endif /* plcLib_h is not defined */
